// use std::net::SocketAddr;

use std::{collections::HashMap, error::Error, fmt::Display};

use tokio::io::BufReader;

use crate::utils::mime::mime_enum::Mime;

static TWO_NEWLINES: u8 = 3;

use super::{
    methods::Method,
    routes::{Data, Uri},
};

type HeaderMap = Vec<String>;
#[derive(Clone)]
pub struct Request<'a> {
    pub uri: Uri<'a>,
    pub headers: HeaderMap,
    pub method: Method,
    // pub connection: ConnectionMeta,
}

// struct ConnectionMeta {
//     remote: Option<SocketAddr>,
//     // certificates
// }

#[derive(Clone, Debug, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum MediaType {
    Json,
    Plain,
    Html,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum ParseErrors {
    NoData,
    BadData,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct ParseFormError {
    pub error: ParseErrors,
}

impl Display for ParseFormError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.error)
    }
}

impl Display for ParseErrors {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParseErrors::NoData => write!(f, "No Data at key"),
            ParseErrors::BadData => write!(f, "Bad Data at key"),
        }
    }
}

impl Error for ParseFormError {}

impl Request<'_> {
    pub fn can_have_body(&self) -> bool {
        match self.method {
            Method::Post | Method::Put | Method::Patch | Method::Delete => true,
            _ => false,
        }
    }
    pub fn mandatory_body(&self) -> bool {
        match self.method {
            Method::Post | Method::Put | Method::Patch => true,
            _ => false,
        }
    }
    pub fn get_get_form_keys<'a>(
        &'a self,
        keys: &'a [&str],
    ) -> Result<HashMap<&str, Result<&str, ParseFormError>>, ParseFormError> {
        let data = if let Some(val) = self.uri.split_once("?") {
            val
        } else {
            return Err(ParseFormError {
                error: ParseErrors::NoData,
            });
        };
        let data = data
            .1
            .split("&")
            .map(|kvp| kvp.split_once("="))
            .collect::<Vec<Option<(&str, &str)>>>();

        let mut values: HashMap<&str, &str> = HashMap::new();
        for kvp in data {
            let kvp = if let Some(kvp) = kvp {
                kvp
            } else {
                continue;
            };
            values.insert(kvp.0, kvp.1);
        }
        let mut response = HashMap::new();
        for key in keys {
            let entry = if let Some(val) = values.get(key) {
                Ok(*val)
            } else {
                Err(ParseFormError {
                    error: ParseErrors::NoData,
                })
            };
            response.insert((*key).into(), entry);
        }
        Ok(response)
    }

    pub fn get_post_data<'a>(
        &'a self,
        keys: &[&'a str],
        data: &Data,
    ) -> Result<HashMap<String, Result<Vec<u8>, ParseFormError>>, ParseFormError> {
        let boundary;
        let post_type = if let Some(val) = self
            .headers
            .iter()
            .find(|header| header.contains("Content-Type: "))
        {
            let content_type = val.trim().strip_prefix("Content-Type: ").unwrap();
            let type_vec = content_type.split(';').collect::<Vec<&str>>();

            boundary = if let Some(bound) = type_vec.iter().find(|part| part.contains(" boundary="))
            {
                bound
                    .strip_prefix(" boundary=")
                    .unwrap()
                    .trim_end()
                    .trim_matches('"')
            } else {
                ""
            };
            if let Ok(mime) = type_vec[0].trim().parse() {
                mime
            } else {
                return Err(ParseFormError {
                    error: ParseErrors::BadData,
                });
            }
        } else {
            return Err(ParseFormError {
                error: ParseErrors::NoData,
            });
        };

        let data = data.buffer.as_slice();
        let mut keymap: HashMap<String, Result<Vec<u8>, ParseFormError>> =
            HashMap::with_capacity(keys.len());

        for key in keys {
            keymap.entry(key.to_string()).or_insert(Err(ParseFormError {
                error: ParseErrors::NoData,
            }));
        }

        match post_type {
            Mime::ApplicationXWwwFormUrlencoded => {
                for kvp in data.split(|byte| *byte == b'&') {
                    let kvp = kvp
                        .split(|byte| *byte == b'=')
                        .map(|list| list.to_vec())
                        .collect::<Vec<Vec<u8>>>();
                    let key = if let Some(kv) = kvp.get(0) {
                        kv
                    } else {
                        return Err(ParseFormError {
                            error: ParseErrors::BadData,
                        });
                    };
                    let key = if let Ok(kv) = String::from_utf8(key.to_vec()) {
                        kv
                    } else {
                        return Err(ParseFormError {
                            error: ParseErrors::BadData,
                        });
                    };
                    let value = kvp.get(1).ok_or(ParseFormError {
                        error: ParseErrors::NoData,
                    });
                    let thing = if let Some(val) = keymap.get_mut(key.as_str()) {
                        val
                    } else {
                        continue;
                    };
                    *thing = match value {
                        Ok(val) => Ok(val.to_vec()),
                        Err(err) => Err(err),
                    }
                }
            }
            Mime::MultipartFormData => {
                let mut temp_bound = "--".to_string();
                temp_bound.push_str(&format!("{boundary}"));
                let end_boundary = format!("{temp_bound}--\r").as_bytes().to_owned();
                temp_bound.push('\r');
                let boundary = temp_bound.as_bytes();
                Request::get_multipart(data, boundary, &end_boundary, &mut keymap);
            }
            _ => {
                return Err(ParseFormError {
                    error: ParseErrors::BadData,
                })
            }
        };
        Ok(keymap)
    }
    fn decode_multipart<'a>(
        data: &Data,
        keys: &[&'a str],
    ) -> HashMap<String, Result<Vec<u8>, ParseFormError>> {
        let mut data = data.buffer.as_ref();
        let mut buf_reader = BufReader::new(&mut data);
        let mut http_request: Vec<String> = Vec::with_capacity(2);
        todo!()
    }
    fn get_multipart_data(
        data: &[u8],
        boundary: &[u8],
        end_boundary: &[u8],
        map: &mut HashMap<String, Result<Vec<u8>, ParseFormError>>,
    ) {
        let parts = data.split(|byte| byte == &b'\n').collect::<Vec<&[u8]>>();
        let mut current_part: Vec<&[u8]> = vec![];
        let mut current_key: Option<String> = None;
        let mut ignore_line = 0;
        for part in parts {
            if part == &[b'\r'] {
                if let Some(_) = current_key {
                    if ignore_line >= TWO_NEWLINES {
                        current_part.push(&[b'\n']);
                        continue;
                    }
                    ignore_line += 1;
                }
                continue;
            }
            if part == end_boundary {
                if let Some(key) = &current_key {
                    let mut part = current_part.join(&b'\n');
                    if part.ends_with(&[b'\r']) {
                        part.pop();
                    }
                    map.insert(key.to_string(), Ok(part));
                }
                break;
            }
            if part == boundary {
                if let Some(key) = &current_key {
                    let mut part = current_part.join(&b'\n');
                    if part.ends_with(&[b'\r']) {
                        part.pop();
                    }
                    map.insert(key.to_string(), Ok(part));
                }
                current_part = vec![];
                current_key = None;
                ignore_line = 0;
                continue;
            }
            if part.starts_with(b"Content-Disposition: form-data; name=") {
                let headers = part
                    .split(|byte| byte == &b';')
                    .filter(|header| !header.is_empty())
                    .collect::<Vec<_>>();
                if headers.len() < 2 {
                    continue;
                }
                let name = headers[1].split(|byte| byte == &b'=').collect::<Vec<_>>();
                if name.len() != 2 {
                    continue;
                }
                let mkey = String::from_utf8_lossy(name[1])
                    .as_ref()
                    .trim_end()
                    .trim_matches('"')
                    .to_owned();

                if map.contains_key::<str>(&mkey) {
                    current_key = Some(mkey.to_owned());
                }
                continue;
            } else if let Some(_) = &current_key {
                current_part.push(part);
            }
        }
    }
}

#[cfg(test)]
mod test {
    use crate::handling::routes::Data;

    use super::Request;

    #[test]
    fn try_post_text() {
        let req = Request {
            uri: "",
            headers: vec!["Content-Type: application/x-www-form-urlencoded".to_string()],
            method: crate::handling::methods::Method::Post,
        };
        let data = Data {
            buffer: b"message=23&message1=24".to_vec(),
            is_complete: true,
        };
        let map = req.get_post_data(&["message", "message1"], &data).unwrap();
        assert_eq!(
            &b"23".to_vec(),
            map.get("message").unwrap().as_ref().unwrap()
        );
        assert_eq!(
            &b"24".to_vec(),
            map.get("message1").unwrap().as_ref().unwrap()
        );
        let req = Request {
            uri: "",
            headers: vec!["Content-Type: multipart/form-data; boundary=\"boundary\"".to_string()],
            method: crate::handling::methods::Method::Post,
        };
        let data = Data {
            buffer: b"--boundary\r
Content-Disposition: form-data; name=\"field1\"\r
\r
value1\r
--boundary\r
Content-Disposition: form-data; name=\"field2\"; filename=\"example.txt\"\r
\r
va\nlue2\r
--boundary--\r
"
            .to_vec(),
            is_complete: true,
        };
        let map = req.get_post_data(&["field1", "field2"], &data).unwrap();

        assert_eq!(
            &b"value1".to_vec(),
            map.get("field1").unwrap().as_ref().unwrap()
        );
        assert_eq!(
            &b"va\nlue2".to_vec(),
            map.get("field2").unwrap().as_ref().unwrap()
        );
    }
}
